
import time
import json
from nevow import (
    rend,
    inevow,
    tags as T,
)
from twisted.web import (
    http,
    html,
)
from twisted.python.filepath import FilePath
from twisted.web.template import (
    Element,
    XMLFile,
    renderer,
    renderElement,
    tags,
)
from allmydata.web.common import (
    getxmlfile,
    get_arg,
    get_root,
    WebError,
    MultiFormatResource,
)
from allmydata.web.operations import ReloadMixin
from allmydata.interfaces import (
    ICheckAndRepairResults,
    ICheckResults,
)
from allmydata.util import (
    base32,
    dictutil,
)


def json_check_counts(r):
    d = {"count-happiness": r.get_happiness(),
         "count-shares-good": r.get_share_counter_good(),
         "count-shares-needed": r.get_encoding_needed(),
         "count-shares-expected": r.get_encoding_expected(),
         "count-good-share-hosts": r.get_host_counter_good_shares(),
         "count-corrupt-shares": len(r.get_corrupt_shares()),
         "list-corrupt-shares": [ (s.get_longname(), base32.b2a(si), shnum)
                                  for (s, si, shnum)
                                  in r.get_corrupt_shares() ],
         "servers-responding": [s.get_longname()
                                for s in r.get_servers_responding()],
         "sharemap": dict([(shareid,
                            sorted([s.get_longname() for s in servers]))
                           for (shareid, servers)
                           in r.get_sharemap().items()]),
         "count-wrong-shares": r.get_share_counter_wrong(),
         "count-recoverable-versions": r.get_version_counter_recoverable(),
         "count-unrecoverable-versions": r.get_version_counter_unrecoverable(),
         }
    return d

def json_check_results(r):
    if r is None:
        # LIT file
        data = {"storage-index": "",
                "results": {"healthy": True},
                }
        return data
    data = {}
    data["storage-index"] = r.get_storage_index_string()
    data["summary"] = r.get_summary()
    data["results"] = json_check_counts(r)
    data["results"]["healthy"] = r.is_healthy()
    data["results"]["recoverable"] = r.is_recoverable()
    return data

def json_check_and_repair_results(r):
    if r is None:
        # LIT file
        data = {"storage-index": "",
                "repair-attempted": False,
                }
        return data
    data = {}
    data["storage-index"] = r.get_storage_index_string()
    data["repair-attempted"] = r.get_repair_attempted()
    data["repair-successful"] = r.get_repair_successful()
    pre = r.get_pre_repair_results()
    data["pre-repair-results"] = json_check_results(pre)
    post = r.get_post_repair_results()
    data["post-repair-results"] = json_check_results(post)
    return data

class ResultsBase(object):
    # self.client must point to the Client, so we can get nicknames and
    # determine the permuted peer order

    def _join_pathstring(self, path):
        if path:
            pathstring = "/".join(self._html(path))
        else:
            pathstring = "<root>"
        return pathstring

    def _render_results(self, ctx, cr):
        assert ICheckResults(cr)
        c = self.client
        sb = c.get_storage_broker()
        r = []
        def add(name, value):
            r.append(tags.li(name + ": ", value))

        add("Report", tags.pre("\n".join(self._html(cr.get_report()))))

        add("Share Counts",
            "need %d-of-%d, have %d" % (cr.get_encoding_needed(),
                                        cr.get_encoding_expected(),
                                        cr.get_share_counter_good()))
        add("Happiness Level", str(cr.get_happiness()))
        add("Hosts with good shares", str(cr.get_host_counter_good_shares()))

        rrr = cr.get_corrupt_shares()

        if cr.get_corrupt_shares():
            badsharemap = []
            for (s, si, shnum) in cr.get_corrupt_shares():
                d = tags.tr(tags.td("sh#%d" % shnum),
                            tags.td(tags.div(s.get_nickname(), class_="nickname"),
                                    tags.div(tags.tt(s.get_name()), class_="nodeid")),)
                badsharemap.append(d)
            add("Corrupt shares",
                tags.table(
                    tags.tr(tags.th("Share ID"),
                            tags.th((tags.div("Nickname"), tags.div("Node ID", class_="nodeid")), class_="nickname-and-peerid")),
                    badsharemap))
        else:
            add("Corrupt shares", "none")

        add("Wrong Shares", str(cr.get_share_counter_wrong()))

        sharemap_data = []
        shares_on_server = dictutil.DictOfSets()

        # FIXME: The two tables below contain nickname-and-nodeid
        # table column markup which is duplicated with each other,
        # introducer.xhtml, and deep-check-results.xhtml. All of these
        # (and any other presentations of nickname-and-nodeid) should be combined.

        for shareid in sorted(cr.get_sharemap().keys()):
            servers = sorted(cr.get_sharemap()[shareid],
                             key=lambda s: s.get_longname())
            for i,s in enumerate(servers):
                shares_on_server.add(s, shareid)
                shareid_s = ""
                if i == 0:
                    shareid_s = str(shareid)
                d = tags.tr(tags.td(shareid_s),
                            tags.td(tags.div(s.get_nickname(), class_="nickname"),
                                    tags.div(tags.tt(s.get_name()), class_="nodeid")))
                sharemap_data.append(d)

        add("Good Shares (sorted in share order)",
            tags.table(tags.tr(tags.th("Share ID"),
                               tags.th(tags.div("Nickname"),
                                       tags.div("Node ID", class_="nodeid"), class_="nickname-and-peerid")),
                       sharemap_data))

        add("Recoverable Versions", str(cr.get_version_counter_recoverable()))
        add("Unrecoverable Versions", str(cr.get_version_counter_unrecoverable()))

        # this table is sorted by permuted order
        permuted_servers = [s
                            for s
                            in sb.get_servers_for_psi(cr.get_storage_index())]

        num_shares_left = sum([len(shareids)
                               for shareids in shares_on_server.values()])
        servermap = []
        for s in permuted_servers:
            shareids = list(shares_on_server.get(s, []))
            shareids.reverse()
            shareids_s = [tags.tt(str(shareid), " ") for shareid in sorted(shareids)]

            d = tags.tr(tags.td(tags.div(s.get_nickname(), class_="nickname"),
                             tags.div(tags.tt(s.get_name()), class_="nodeid")),
                        tags.td(shareids_s), )
            servermap.append(d)
            num_shares_left -= len(shareids)
            if not num_shares_left:
                break

        add("Share Balancing (servers in permuted order)",
            tags.table(tags.tr(tags.th(tags.div("Nickname"),
                                    tags.div("Node ID", class_="nodeid"), class_="nickname-and-peerid"),
                            tags.th("Share IDs")),
                       servermap))

        return tags.ul(r)

    def _html(self, s):
        if isinstance(s, (str, unicode)):
            return html.escape(s)
        assert isinstance(s, (list, tuple))
        return [html.escape(w) for w in s]

    def want_json(self, ctx):
        output = get_arg(inevow.IRequest(ctx), "output", "").lower()
        if output.lower() == "json":
            return True
        return False

    def _render_si_link(self, ctx, storage_index):
        si_s = base32.b2a(storage_index)
        req = inevow.IRequest(ctx)
        ophandle = req.prepath[-1]
        target = "%s/operations/%s/%s" % (get_root(ctx), ophandle, si_s)
        output = get_arg(ctx, "output")
        if output:
            target = target + "?output=%s" % output
        return tags.a(si_s, href=target)


class LiteralCheckResultsRenderer(MultiFormatResource, ResultsBase):

    formatArgument = "output"

    def __init__(self, client):
        super(LiteralCheckResultsRenderer, self).__init__()
        self.client = client

    def render_HTML(self, req):
        return renderElement(req, LiteralCheckResultsElement())

    def render_JSON(self, req):
        req.setHeader("content-type", "text/plain")
        data = json_check_results(None)
        return json.dumps(data, indent=1) + "\n"


class LiteralCheckResultsElement(Element):

    loader = XMLFile(FilePath(__file__).sibling("literal-check-results.xhtml"))

    def __init__(self):
        super(LiteralCheckResultsElement, self).__init__()

    @renderer
    def return_to(self, req, tag):
        return_to = get_arg(req, "return_to", None)
        if return_to:
            return tags.div(tags.a("Return to file.", href=return_to))
        return ""


class CheckerBase(object):

    # def renderHTTP(self, ctx):
    #     if self.want_json(ctx):
    #         return self.json(ctx)
    #     return rend.Page.renderHTTP(self, ctx)

    @renderer
    def storage_index(self, req, tag):
        return self.r.get_storage_index_string()

    @renderer
    def return_to(self, req, tag):
        # req = inevow.IRequest(ctx)
        return_to = get_arg(req, "return_to", None)
        if return_to:
            return tags.div(tags.a("Return to file/directory.", href=return_to))
        return ""


class CheckResultsRenderer(MultiFormatResource):

    formatArgument = "output"

    def __init__(self, client, results):
        """
        TODO: update these with the correct values.
        :param allmydata.test.no_network._NoNetworkClient client:
        :param allmydata.check_results.CheckResults results:
        """
        super(CheckResultsRenderer, self).__init__()
        self.client = client
        self.r = ICheckResults(results)

    def render_HTML(self, req):
        return renderElement(req, CheckResultsRendererElement(self.client, self.r))

    def render_JSON(self, req):
        req.setHeader("content-type", "text/plain")
        data = json_check_results(self.r)
        return json.dumps(data, indent=1) + "\n"


class CheckResultsRendererElement(Element, CheckerBase, ResultsBase):

    loader = XMLFile(FilePath(__file__).sibling("check-results.xhtml"))

    def __init__(self, client, r):
        super(CheckResultsRendererElement, self).__init__()
        self.client = client
        # TODO: use a better name
        self.r = r

    @renderer
    def summary(self, req, tag):
        results = []
        # XXX: how did `data` get populated?
        if self.r.is_healthy():
            results.append("Healthy")
        elif self.r.is_recoverable():
            results.append("Not Healthy!")
        else:
            results.append("Not Recoverable!")
        results.append(" : ")
        # TODO: what is self._html?
        # results.append(self._html(self.r.get_summary()))
        results.append(self.r.get_summary())
        return tag(results)

    @renderer
    def repair(self, req, tag):
        if self.r.is_healthy():
            return ""

        #repair = T.form(action=".", method="post",
        #                enctype="multipart/form-data")[
        #    T.fieldset[
        #    T.input(type="hidden", name="t", value="check"),
        #    T.input(type="hidden", name="repair", value="true"),
        #    T.input(type="submit", value="Repair"),
        #    ]]
        #return ctx.tag[repair]

        return "" # repair button disabled until we make it work correctly,
                  # see #622 for details

    @renderer
    def results(self, req, tag):
        cr = self._render_results(req, self.r)
        return tag(cr)


class CheckAndRepairResultsRenderer(MultiFormatResource):

    formatArgument = "output"

    def __init__(self, client, results):
        # TODO: document params
        super(CheckAndRepairResultsRenderer, self).__init__()
        self.client = client
        # TODO: use a better name
        self.r = None
        if results:
            self.r = ICheckAndRepairResults(results)

    def render_HTML(self, req):
        elem = CheckAndRepairResultsRendererElement(self.r)
        return renderElement(req, elem)

    def render_JSON(self, req):
        req.setHeader("content-type", "text/plain")
        data = json_check_and_repair_results(self.r)
        return json.dumps(data, indent=1) + "\n"


class CheckAndRepairResultsRendererElement(Element, CheckerBase, ResultsBase):

    loader = XMLFile(FilePath(__file__).sibling("check-and-repair-results.xhtml"))

    def __init__(self, r):
        super(CheckAndRepairResultsRendererElement, self).__init__()
        self.r = r

    @renderer
    def summary(self, req, tag):
        cr = self.r.get_post_repair_results()
        results = []
        if cr.is_healthy():
            results.append("Healthy")
        elif cr.is_recoverable():
            results.append("Not Healthy!")
        else:
            results.append("Not Recoverable!")
        results.append(" : ")
        results.append(self._html(cr.get_summary()))
        return tag(results)

    @renderer
    def repair_results(self, req, tag):
        if self.r.get_repair_attempted():
            if self.r.get_repair_successful():
                return tag("Repair successful")
            else:
                return tag("Repair unsuccessful")
        return tag("No repair necessary")

    @renderer
    def post_repair_results(self, req, tag):
        cr = self._render_results(req, self.r.get_post_repair_results())
        return tag(tags.div("Post-Repair Checker Results:"), cr)

    @renderer
    def maybe_pre_repair_results(self, req, tag):
        if self.r.get_repair_attempted():
            cr = self._render_results(req, self.r.get_pre_repair_results())
            return tag(tags.div("Pre-Repair Checker Results:"), cr)
        return ""


class DeepCheckResultsRenderer(MultiFormatResource):

    formatArgument = "output"

    def __init__(self, client, monitor):
        # TODO: document params
        super(DeepCheckResultsRenderer, self).__init__()
        self.client = client
        self.monitor = monitor

    def getChild(self, name, req):
        if not name:
            return self
        # /operation/$OPHANDLE/$STORAGEINDEX provides detailed information
        # about a specific file or directory that was checked
        si = base32.a2b(name)
        r = self.monitor.get_status()
        try:
            return CheckResultsRenderer(self.client,
                                        r.get_results_for_storage_index(si))
        except KeyError:
            raise WebError("No detailed results for SI %s" % html.escape(name),
                           http.NOT_FOUND)

    def render_HTML(self, req):
        elem = DeepCheckResultsRendererElement(self.monitor)
        return renderElement(req, elem)

    # def renderHTTP(self, ctx):
    #     if self.want_json(ctx):
    #         return self.json(ctx)
    #     return rend.Page.renderHTTP(self, ctx)

    def render_JSON(self, req):
        # inevow.IRequest(ctx).setHeader("content-type", "text/plain")
        req.setHeader("content-type", "text/plain")
        data = {}
        data["finished"] = self.monitor.is_finished()
        res = self.monitor.get_status()
        data["root-storage-index"] = res.get_root_storage_index_string()
        c = res.get_counters()
        data["count-objects-checked"] = c["count-objects-checked"]
        data["count-objects-healthy"] = c["count-objects-healthy"]
        data["count-objects-unhealthy"] = c["count-objects-unhealthy"]
        data["count-corrupt-shares"] = c["count-corrupt-shares"]
        data["list-corrupt-shares"] = [ (s.get_longname(),
                                         base32.b2a(storage_index),
                                         shnum)
                                        for (s, storage_index, shnum)
                                        in res.get_corrupt_shares() ]
        data["list-unhealthy-files"] = [ (path_t, json_check_results(r))
                                         for (path_t, r)
                                         in res.get_all_results().items()
                                         if not r.is_healthy() ]
        data["stats"] = res.get_stats()
        return json.dumps(data, indent=1) + "\n"


class DeepCheckResultsRendererElement(Element, ResultsBase, ReloadMixin):

    loader = XMLFile(FilePath(__file__).sibling("deep-check-results.xhtml"))

    def __init__(self, monitor):
        super(DeepCheckResultsRendererElement, self).__init__()
        self.monitor = monitor

    @renderer
    def root_storage_index(self, req, tag):
        return self.monitor.get_status().get_root_storage_index_string()

    @renderer
    def objects_checked(self, req, tag):
        return self.monitor.get_status().get_counters()["count-objects-checked"]

    @renderer
    def objects_healthy(self, req, tag):
        return self.monitor.get_status().get_counters()["count-objects-healthy"]

    @renderer
    def objects_unhealthy(self, req, tag):
        return self.monitor.get_status().get_counters()["count-objects-unhealthy"]

    @renderer
    def objects_unrecoverable(self, req, tag):
        return self.monitor.get_status().get_counters()["count-objects-unrecoverable"]

    @renderer
    def count_corrupt_shares(self, req, tag):
        return self.monitor.get_status().get_counters()["count-corrupt-shares"]

    @renderer
    def problems_p(self, req, tag):
        c = self.monitor.get_status().get_counters()
        if c["count-objects-unhealthy"]:
            return tag
        return ""

    # TODO: use SlotsSequenceElement to render this.
    @renderer
    def problems(self, req, tag):
        all_objects = self.monitor.get_status().get_all_results()
        for path in sorted(all_objects.keys()):
            cr = all_objects[path]
            assert ICheckResults.providedBy(cr)
            if not cr.is_healthy():
                yield path, cr

    @renderer
    def render_problem(self, req, tag):
        # TODO: deal with this.
        path, cr = data
        summary_text = ""
        summary = cr.get_summary()
        if summary:
            summary_text = ": " + summary
        summary_text += " [SI: %s]" % cr.get_storage_index_string()
        return tag(self._join_pathstring(path), self._html(summary_text))

    @renderer
    def servers_with_corrupt_shares_p(self, req, tag):
        if self.monitor.get_status().get_counters()["count-corrupt-shares"]:
            return tag
        return ""

    # TODO: use SlotsSequenceElement to render this.
    @renderer
    def servers_with_corrupt_shares(self, ctx, data):
        servers = [s
                   for (s, storage_index, sharenum)
                   in self.monitor.get_status().get_corrupt_shares()]
        servers.sort(key=lambda s: s.get_longname())
        return servers

    @renderer
    def server_problem(self, req, tag):
    # def server_problem(self, ctx, server):
        # TODO: where do `server` come from now?
        data = [server.get_name()]
        nickname = server.get_nickname()
        if nickname:
            data.append(" (%s)" % self._html(nickname))
        return tag(data)

    @renderer
    def corrupt_shares_p(self, req, tag):
        if self.monitor.get_status().get_counters()["count-corrupt-shares"]:
            return tag
        return ""

    # TODO: Probably should use SlotsSequenceElement to render this.
    @renderer
    def corrupt_shares(self, req, tag):
        return self.monitor.get_status().get_corrupt_shares()

    @renderer
    def share_problem(self, req, tag):
    # def render_share_problem(self, req, tag):
        server, storage_index, sharenum = data
        nickname = server.get_nickname()
        tag.fillSlots("serverid", server.get_name())
        if nickname:
            tag.fillSlots("nickname", self._html(nickname))
        tag.fillSlots("si", self._render_si_link(ctx, storage_index))
        tag.fillSlots("shnum", str(sharenum))
        return tag

    @renderer
    def return_to(self, req, tag):
        # req = inevow.IRequest(ctx)
        return_to = get_arg(req, "return_to", None)
        if return_to:
            return tags.div(tags.a("Return to file/directory.", href=return_to))
        return ""

    # TODO: use SlotsSequenceElement to render this.
    @renderer
    def all_objects(self, ctx, data):
        r = self.monitor.get_status().get_all_results()
        for path in sorted(r.keys()):
            yield (path, r[path])

    @renderer
    def object(self, ctx, data):
    # def render_object(self, ctx, data):
        # TODO: figure `data` out
        path, r = data
        tag.fillSlots("path", self._join_pathstring(path))
        tag.fillSlots("healthy", str(r.is_healthy()))
        tag.fillSlots("recoverable", str(r.is_recoverable()))
        storage_index = r.get_storage_index()
        tag.fillSlots("storage_index", self._render_si_link(req, storage_index))
        tag.fillSlots("summary", self._html(r.get_summary()))
        return tag

    @renderer
    def runtime(self, req, tag):
        # req = inevow.IRequest(ctx)
        runtime = time.time() - req.processing_started_timestamp
        return tag("runtime: %s seconds" % runtime)


class DeepCheckAndRepairResultsRenderer(MultiFormatResource):

    formatArgument = "output"

    def __init__(self, client, monitor):
        self.client = client
        self.monitor = monitor

    def getChild(self, name, req):
        if not name:
            return self
        # /operation/$OPHANDLE/$STORAGEINDEX provides detailed information
        # about a specific file or directory that was checked
        si = base32.a2b(name)
        s = self.monitor.get_status()
        try:
            results = s.get_results_for_storage_index(si)
            return CheckAndRepairResultsRenderer(self.client, results)
        except KeyError:
            raise WebError("No detailed results for SI %s" % html.escape(name),
                           http.NOT_FOUND)

    def render_HTML(self, req):
        elem = DeepCheckAndRepairResultsRendererElement(self.monitor)
        return renderElement(req, elem)

    # def renderHTTP(self, ctx):
    #     if self.want_json(ctx):
    #         return self.json(ctx)
    #     return rend.Page.renderHTTP(self, ctx)

    def render_JSON(self, req):
        # inevow.IRequest(ctx).setHeader("content-type", "text/plain")
        req.setHeader("content-type", "text/plain")
        res = self.monitor.get_status()
        data = {}
        data["finished"] = self.monitor.is_finished()
        data["root-storage-index"] = res.get_root_storage_index_string()
        c = res.get_counters()
        data["count-objects-checked"] = c["count-objects-checked"]

        data["count-objects-healthy-pre-repair"] = c["count-objects-healthy-pre-repair"]
        data["count-objects-unhealthy-pre-repair"] = c["count-objects-unhealthy-pre-repair"]
        data["count-objects-healthy-post-repair"] = c["count-objects-healthy-post-repair"]
        data["count-objects-unhealthy-post-repair"] = c["count-objects-unhealthy-post-repair"]

        data["count-repairs-attempted"] = c["count-repairs-attempted"]
        data["count-repairs-successful"] = c["count-repairs-successful"]
        data["count-repairs-unsuccessful"] = c["count-repairs-unsuccessful"]

        data["count-corrupt-shares-pre-repair"] = c["count-corrupt-shares-pre-repair"]
        data["count-corrupt-shares-post-repair"] = c["count-corrupt-shares-pre-repair"]

        data["list-corrupt-shares"] = [ (s.get_longname(),
                                         base32.b2a(storage_index),
                                         shnum)
                                        for (s, storage_index, shnum)
                                        in res.get_corrupt_shares() ]

        remaining_corrupt = [ (s.get_longname(), base32.b2a(storage_index),
                               shnum)
                              for (s, storage_index, shnum)
                              in res.get_remaining_corrupt_shares() ]
        data["list-remaining-corrupt-shares"] = remaining_corrupt

        unhealthy = [ (path_t,
                       json_check_results(crr.get_pre_repair_results()))
                      for (path_t, crr)
                      in res.get_all_results().items()
                      if not crr.get_pre_repair_results().is_healthy() ]
        data["list-unhealthy-files"] = unhealthy
        data["stats"] = res.get_stats()
        return json.dumps(data, indent=1) + "\n"


class DeepCheckAndRepairResultsRendererElement(Element, ResultsBase, ReloadMixin):

    loader = XMLFile(FilePath(__file__).sibling("deep-check-and-repair-results.xhtml"))

    def __init__(self, monitor):
        # TODO: document params
        super(DeepCheckAndRepairResultsRendererElement, self).__init__()
        self.monitor = monitor

    @renderer
    def root_storage_index(self, req, tag):
        return self.monitor.get_status().get_root_storage_index_string()

    @renderer
    def objects_checked(self, req, tag):
        return self.monitor.get_status().get_counters()["count-objects-checked"]

    @renderer
    def objects_healthy(self, req, tag):
        return self.monitor.get_status().get_counters()["count-objects-healthy-pre-repair"]

    @renderer
    def objects_unhealthy(self, req, tag):
        return self.monitor.get_status().get_counters()["count-objects-unhealthy-pre-repair"]

    @renderer
    def corrupt_shares(self, req, tag):
        return self.monitor.get_status().get_counters()["count-corrupt-shares-pre-repair"]

    @renderer
    def repairs_attempted(self, req, tag):
        return self.monitor.get_status().get_counters()["count-repairs-attempted"]

    @renderer
    def repairs_successful(self, req, tag):
        return self.monitor.get_status().get_counters()["count-repairs-successful"]

    @renderer
    def repairs_unsuccessful(self, req, tag):
        return self.monitor.get_status().get_counters()["count-repairs-unsuccessful"]

    @renderer
    def objects_healthy_post(self, req, tag):
        return self.monitor.get_status().get_counters()["count-objects-healthy-post-repair"]

    @renderer
    def objects_unhealthy_post(self, req, tag):
        return self.monitor.get_status().get_counters()["count-objects-unhealthy-post-repair"]

    @renderer
    def corrupt_shares_post(self, req, tag):
        return self.monitor.get_status().get_counters()["count-corrupt-shares-post-repair"]

    @renderer
    def pre_repair_problems_p(self, req, tag):
        c = self.monitor.get_status().get_counters()
        if c["count-objects-unhealthy-pre-repair"]:
            return tag
        return ""

    # TODO: use SlotsSequenceElement
    @renderer
    def pre_repair_problems(self, req, tag):
        all_objects = self.monitor.get_status().get_all_results()
        for path in sorted(all_objects.keys()):
            r = all_objects[path]
            assert ICheckAndRepairResults.providedBy(r)
            cr = r.get_pre_repair_results()
            if not cr.is_healthy():
                yield path, cr

    @renderer
    def problem(self, req, tag):
        # TODO: figure this out
        # path, cr = data
        return tag(self._join_pathstring(path), ": ",
                   self._html(cr.get_summary()))

    @renderer
    def post_repair_problems_p(self, req, tag):
        c = self.monitor.get_status().get_counters()
        if (c["count-objects-unhealthy-post-repair"]
            or c["count-corrupt-shares-post-repair"]):
            return tag
        return ""

    # TODO: use SlotsSequenceElement
    @renderer
    def post_repair_problems(self, req, tag):
        all_objects = self.monitor.get_status().get_all_results()
        for path in sorted(all_objects.keys()):
            r = all_objects[path]
            assert ICheckAndRepairResults.providedBy(r)
            cr = r.get_post_repair_results()
            if not cr.is_healthy():
                yield path, cr

    @renderer
    def servers_with_corrupt_shares_p(self, req, tag):
        if self.monitor.get_status().get_counters()["count-corrupt-shares-pre-repair"]:
            return tag
        return ""

    @renderer
    def servers_with_corrupt_shares(self, req, tag):
        return [] # TODO

    @renderer
    def server_problem(self, req, tag):
        pass

    @renderer
    def remaining_corrupt_shares_p(self, req, tag):
        if self.monitor.get_status().get_counters()["count-corrupt-shares-post-repair"]:
            return tag
        return ""

    @renderer
    def post_repair_corrupt_shares(self, req, tag):
        return [] # TODO

    @renderer
    def render_share_problem(self, req, tag):
        pass

    @renderer
    def return_to(self, req, tag):
        # req = inevow.IRequest(ctx)
        return_to = get_arg(req, "return_to", None)
        if return_to:
            return tags.div(tags.a("Return to file/directory.", href=return_to))
        return ""

    # TODO: use SlotsSequenceElement
    @renderer
    def all_objects(self, req, tag):
        r = self.monitor.get_status().get_all_results()
        for path in sorted(r.keys()):
            yield (path, r[path])

    @renderer
    def object(self, req, tag):
        # TODO: figure this out
        # path, r = data
        tag.fillSlots("path", self._join_pathstring(path))
        tag.fillSlots("healthy_pre_repair",
                      str(r.get_pre_repair_results().is_healthy()))
        tag.fillSlots("recoverable_pre_repair",
                      str(r.get_pre_repair_results().is_recoverable()))
        tag.fillSlots("healthy_post_repair",
                      str(r.get_post_repair_results().is_healthy()))
        storage_index = r.get_storage_index()
        tag.fillSlots("storage_index",
                      self._render_si_link(ctx, storage_index))
        tag.fillSlots("summary",
                      self._html(r.get_pre_repair_results().get_summary()))
        return tag

    @renderer
    def runtime(self, req, tag):
        # req = inevow.IRequest(ctx)
        runtime = time.time() - req.processing_started_timestamp
        return tag("runtime: %s seconds" % runtime)
